#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
#include <png.h>

#include <GL/gl.h>
#include <SDL/SDL.h>

#define min(x, y) ((x) > (y) ? (y) : (x))
#define max(x, y) ((x) > (y) ? (x) : (y))

float mutiply_and_sum_vectors(float vector1[3], float vector2[3]) {
    float sum = 0;
    unsigned i;
    for (i = 0; i < 3; i++) {
	sum += vector1[i] * vector2[i];
    }
    return sum;
}

void multiply_vector_with_matrix(float vector[3], float matrix[9]) {
    unsigned i;
    float tmp[3];
    for (i = 0; i < 3; i++) {
	tmp[i] = mutiply_and_sum_vectors(vector, matrix + (3 * i));
    }
    memcpy(vector, tmp, 3 * sizeof(float));
}

float vector_length(float a[3]) {
    return sqrt(a[0] * a[0] + a[1] * a[1] + a[2] * a[2]);
}

void rotate_vector(float vector[3], float axis[3], float angle) {
    float len = vector_length(axis);
    if (len > 0.0) {
	float c = cos(-angle);
	float s = sin(-angle);
	float x = axis[0] / len;
	float y = axis[1] / len;
	float z = axis[2] / len;
	float matrix[9] = { x*x+(1-x*x)*c, x*y*(1-c)-z*s, x*z*(1-c)+y*s, 
			     x*y*(1-c)+z*s, y*y+(1-y*y)*c, y*z*(1-c)-x*s,
			     x*z*(1-c)-y*s, y*z*(1-c)+x*s, z*z+(1-z*z)*c };
	multiply_vector_with_matrix(vector, matrix);
    }
}

float dot_product(float a[3], float b[3]) {
    return a[0] * b[0] + a[1] * b[1] + a[2] * b[2];
}

float get_angle(float a[3], float b[3]) {
    float x = dot_product(a, b) / (vector_length(a) * vector_length (b));
    return acos(min(1.0, max(-1.0, x)));
}

#define X 0
#define Y 1
#define Z 2

#define X_AXIS 0
#define Y_AXIS 3
#define Z_AXIS 6

float x_axis[] = { 1.0, 0.0, 0.0 };
float y_axis[] = { 0.0, 1.0, 0.0 };
float z_axis[] = { 0.0, 0.0, 1.0 };

#define HALF_PI 1.5707963267948966

void rotate_space(float vector[3], float space[9]) {
    float new_x_axis[] = { 1.0, 0.0, 0.0 };
    float y_deviance = get_angle(space + Y_AXIS, y_axis);
    float z_deviance = -atan2(space[Y_AXIS + X], space[Y_AXIS + Z]);

    float c = cos(y_deviance);
    float s = sin(y_deviance);
    float r_x[9] = { 1, 0, 0, 0, c,-s, 0, s, c };

    c = cos(z_deviance);
    s = sin(z_deviance);
    float r_y[9] = { c, 0,-s, 0, 1, 0, s, 0, c };

    multiply_vector_with_matrix(vector, r_x);
    multiply_vector_with_matrix(vector, r_y);

    multiply_vector_with_matrix(new_x_axis, r_x);
    multiply_vector_with_matrix(new_x_axis, r_y);

    float tmp[3];
    memcpy(tmp, space + Y_AXIS, 3 * sizeof(float));
    rotate_vector(vector,
		  tmp,
		  get_angle(new_x_axis, space + X_AXIS)
  		  *
		  (get_angle(new_x_axis, space + Z_AXIS) < HALF_PI ? -1 : +1));
}

unsigned set_opengl_texture(char *buf, int width, int height, int channels) {
    unsigned texture = 0;
    GLenum format = (channels == 3 ? GL_RGB : GL_RGBA);
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexImage2D(GL_TEXTURE_2D, 0, channels, width, height,
		 0, format, GL_UNSIGNED_BYTE, buf);
    glBindTexture(GL_TEXTURE_2D, 0);
    return texture;
}

#define PNG_BYTES_TO_CHECK 4
int check_if_png(FILE *fp) {
    char buf[PNG_BYTES_TO_CHECK];
    return (fread(buf, 1, PNG_BYTES_TO_CHECK, fp) == PNG_BYTES_TO_CHECK)
	&& !png_sig_cmp(buf, (png_size_t) 0, PNG_BYTES_TO_CHECK);
}

unsigned char *png_data = NULL;
int png_channels;
int png_w;
int png_h;

void get_raw_image(png_structp png_ptr, png_infop info_ptr) {
    unsigned bytes_per_row = png_get_rowbytes(png_ptr, info_ptr);
    unsigned height = png_get_image_height(png_ptr, info_ptr);        
    png_bytepp row_data = png_get_rows(png_ptr, info_ptr);
    png_data = malloc(bytes_per_row * height);
    unsigned i;
    for (i = 0; i < height; i++) {
	memcpy(png_data + i * bytes_per_row, row_data[i], bytes_per_row);
    }
}

unsigned char *load_png(const char *name) {
    png_infop info_ptr;
    png_infop end_info;
    png_structp png_ptr = NULL;    
    FILE *png_fp = fopen(name, "rb");

    if (!png_fp || !check_if_png(png_fp)) {
	goto png_end;
    }

    png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
    if (!png_ptr) {
	goto png_end;
    }
    
    if (!(info_ptr = png_create_info_struct(png_ptr))) {
        png_destroy_read_struct(&png_ptr, NULL, NULL);
	goto png_end;
    }
    
    if (!(end_info = png_create_info_struct(png_ptr))) {
        png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	goto png_end;
    }

    png_init_io(png_ptr, png_fp);
    png_set_sig_bytes(png_ptr, PNG_BYTES_TO_CHECK);
    png_read_png(png_ptr, info_ptr, 0, NULL);

    get_raw_image(png_ptr, info_ptr);
    png_w = png_get_image_width(png_ptr, info_ptr);
    png_h = png_get_image_height(png_ptr, info_ptr);
    png_channels = png_get_channels(png_ptr, info_ptr);

    png_destroy_read_struct(&png_ptr, &info_ptr, &end_info);
  png_end:
    if (png_fp) fclose(png_fp);
    return png_data;
}

void free_png(void) {
    if (png_data) {
	free(png_data);
	png_data = NULL;
    }    
}

unsigned load_texture(const char *name) {
    unsigned texture = 0;
    
    if (load_png(name)) {    
	texture = set_opengl_texture(png_data, png_w, png_h, png_channels);
    }
    free_png();

    if (!texture) printf("ERROR: bad %s\n", name);
    return texture;
}

void delete_texture(unsigned texture) {
    glDeleteTextures(1, &texture);
}

unsigned screen_w = 640;
unsigned screen_h = 480;

float get_w(void) {
    return screen_w;
}

float get_h(void) {
    return screen_h;
}

#define BYTES_PER_PIXEL 3

static unsigned screenBufferWidth(void) {
    return screen_w * BYTES_PER_PIXEL;
}

static unsigned screenBufferSize(void) {
    return screen_h * screenBufferWidth();
}

static unsigned bufferY(unsigned offset) {
    return offset / screenBufferWidth();
}

static unsigned bufferX(unsigned offset) {
    return (offset % screenBufferWidth()) / BYTES_PER_PIXEL;
}

static unsigned bufferOffset(unsigned x, unsigned y) {
    return (y * screenBufferWidth() + x * BYTES_PER_PIXEL);
}

static void copy_screen(unsigned char *buf) {
    int i, j = 0;
    unsigned char tmp[screenBufferWidth()];
    glReadPixels(0, 0, screen_w, screen_h, GL_RGB, GL_UNSIGNED_BYTE, buf);
    for (i = (screen_h - 1); i >= screen_h / 2; i--) {
	memcpy(tmp, buf + j, screenBufferWidth());
	memcpy(buf + j, buf + (i * screenBufferWidth()), screenBufferWidth());
	memcpy(buf + (i * screenBufferWidth()), tmp, screenBufferWidth());
	j += screenBufferWidth();
    }
}

void save_screen_buffer(const char *file_name, unsigned char *buf) {
    int fd = open(file_name,  O_CREAT | O_WRONLY, 0644);
    if (fd != -1) {
	char header[256];
	unsigned char ptr[screenBufferSize()];
	unsigned len = sprintf(header, "P6\n%u %u\n255\n", screen_w, screen_h);

	if (!buf) {
	    copy_screen(ptr);
	    buf = ptr;
	}

	write(fd, header, len);
	write(fd, buf, screenBufferSize());
	close(fd);
    }
    else {
	printf("ERROR: can not write %s\n", file_name);
    }
}

void save_screen(const char *file_name) {
    save_screen_buffer(file_name, NULL);
}

SDL_Surface *screen = NULL;

unsigned flags = SDL_OPENGL | SDL_RESIZABLE | SDL_HWSURFACE;

static void init_video_mode(void) {
    screen = SDL_SetVideoMode(screen_w, screen_h, 32, flags);    
    glViewport(0, 0, screen_w, screen_h);
}

static void resize_window(unsigned w, unsigned h) {
    screen_w = w;
    screen_h = h;
    SDL_FreeSurface(screen);
    init_video_mode();
}

void swap_buffers(void) {
    SDL_GL_SwapBuffers();
}

void exit_graphics(void) {
    if (screen) {
	SDL_FreeSurface(screen);
	screen = NULL;
	SDL_Quit();
    }
}

void init_graphics(void) {
    if (!screen) {
	SDL_Init(SDL_INIT_VIDEO);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	init_video_mode();
	SDL_WM_SetCaption("sandbox", NULL);
	SDL_ShowCursor(SDL_DISABLE);
    }
}
 
unsigned some_mouse_button = 0;

void event_loop(void (*draw_scene)(void),		
		int (*key_change)(int, int),
		void (*mouse_move)(int, int, int, int, int)) {
    int done = 0;
    SDL_Event event;
    some_mouse_button = 0;
    while(!done && SDL_WaitEvent(&event)) {
	do {
	    switch(event.type) {	    
	    case SDL_VIDEORESIZE:
		resize_window(event.resize.w, event.resize.h);
		break;
	    case SDL_KEYDOWN:
	    case SDL_KEYUP:
		if (key_change(event.key.state == SDL_PRESSED, 
			       event.key.keysym.sym)) {
		    done = 1;
		}
		break;
	    case SDL_MOUSEBUTTONDOWN:
		some_mouse_button = 1;
		break;
	    case SDL_MOUSEBUTTONUP:
		some_mouse_button = 0;
		break;
	    case SDL_MOUSEMOTION:
		mouse_move(event.motion.x, 
			   event.motion.y,
			   event.motion.xrel,
			   event.motion.yrel, 
			   some_mouse_button);
		break;	
	    case SDL_QUIT:
		done = 1;
		break;            
	    default:
		break;
	    } 
	}
	while (SDL_PeepEvents(&event, 1, SDL_GETEVENT, SDL_ALLEVENTS));
	draw_scene();
    }
}

int crandom(unsigned modulo) {
#ifdef unix
    return random() % modulo;
#else
    return rand() % modulo;
#endif
}

void cseed(unsigned seed) {
#ifdef unix
    srandom(seed);
#else
    srand(seed);
#endif
}

void emit_char(int num) {
    char buf[] = { (char) num };
    write(1, buf, 1);
}
